from distutils.core import setup
setup(
  name = 'autodiffad',         
  packages = ['autodiffad'],   
  version = '1.0.3',      
  license='MIT',        
  description = 'Automatic Differentiation',   
  author = 'cs107-zero-day-exploit',                   
  author_email = 'yichunyao@hsph.harvard.edu',      
  url = 'https://github.com/cs107-zero-day-exploit/cs107-FinalProject',   
  download_url = 'https://github.com/cs107-zero-day-exploit/cs107-FinalProject/archive/refs/tags/v1.0.3.tar.gz',
  keywords = ['automatic', 'differentiation', 'forward mode', 'reverse mode'],   
  install_requires=[            
          'numpy',
          'pytest',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)



# -*- coding: utf-8 -*-

import numpy as np
import builtins as b

class ReverseMode():
    
    # The ReverseMode class is the equivalent of the DualNumber Class in forward mode
    # In this class the overloading of all operations (add, mul, etc.) is implemented
    # as well as the tree structure and the recursive traversal of the tree through its children
    
    def __init__(self, value):
        """
        Initializes ReverseMode object with a value that was passed in
        Inputs: Scalar or list or numpy array
        Returns: A ReverseMode object initialized with the value that is passed in

        Example:
        >>> x = ReverseMode([2, 3])
        >>> x.val
        array([2, 3])
        """
        if isinstance(value, (int, float)):
            self.value = np.array([value])
        elif isinstance(value, list):
            self.value = np.array(value)
        elif isinstance(value, np.ndarray):
            self.value = value
        else:
            raise TypeError("Enter a valid type (float, int, list or np.ndarray).")
        self.child = []
        self.gradient_coef = None

    def reset_gradient(self):
        """
        Sets the gradient of all its children to None
        Inputs: None
        Returns: None

        Example:
        >>> w = ReverseMode([1,2])
        >>> x = ReverseMode([3,4])
        >>> y = ReverseMode([5,6])
        >>> z = (3*x-w**2.5)/(x+y**w)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.25      , 0.07103553])
        >>> x.reset_gradient()
        >>> grads = []
        >>> for _, child in x.child:
        ...     grads.append(child.gradient_coef)
        ...     for _, child2 in child.child:
        ...         grads.append(child2.gradient_coef)
        ...
        >>> print(grads)
        [None, None, None, None]
        """
        self.gradient_coef = None
        for _, child in self.child:
            child.reset_gradient()

    def get_derivative(self):
        """
        Calculates the gradient with respect to the ReverseMode instance
        Inputs: None
        Returns: The gradient value (float)

        Example:
        >>> w = ReverseMode([1,2])
        >>> x = ReverseMode([3,4])
        >>> y = ReverseMode([5,6])
        >>> z = (3*x-w**2.5)/(x+y**w)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.25      , 0.07103553])
        """
        if self.gradient_coef is None:
            self.gradient_coef = b.sum(
                weight * child.get_derivative() for weight, child in self.child
            )
        return self.gradient_coef

    def __add__(self, other):
        """
        Overloads the addition operation
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the addition operation
        perform

        Example:
        >>> x = ReverseMode([1,2])
        >>> y = ReverseMode([3,4])
        >>> z = x + y
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1., 1.])
        >>>
        >>> x = ReverseMode([5,6])
        >>> z = x + 3
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1., 1.])

        """
        if isinstance(other, int) or isinstance(other, float):
            other = ReverseMode([other]*len(self.value))
        z = ReverseMode(self.value + other.value)
        one_array = np.ones(self.value.shape)
        self.child.append((one_array, z)) # weight = dz/dself = 1
        other.child.append((one_array, z)) # weight = dz/dother = 1
        return z

    def __radd__(self, other):
        """
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the addition operation
        performed between the argument that was passed in and the ReverseMode object

        Example:
        >>> x = ReverseMode([1.5,6.3])
        >>> z = 1 + x
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1., 1.])
        """
        return self + other

    def __sub__(self, other):
        """
        Overloads the subtraction operation
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the subtraction operation
        performed between the ReverseMode object and the argument that was passed in

        Example:
        >>> x = ReverseMode([1,2])
        >>> y = ReverseMode([3,4])
        >>> z = x - y
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1., 1.])
        >>>
        >>> x = ReverseMode([10,20])
        >>> z = x - 3
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1., 1.])
        """
        if isinstance(other, int) or isinstance(other, float):
            other = ReverseMode([other]*len(self.value))
        z = ReverseMode(self.value - other.value)
        self.child.append((np.ones(self.value.shape), z)) # weight = dz/dself = 1
        other.child.append((-np.ones(self.value.shape), z)) # weight = dz/dother = -1
        return z

    def __rsub__(self, other):
        """
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the subtraction operation
        performed between the ReverseMode object and the argument that was passed in

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = 2 - x
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-1., -1.])
        """
        if isinstance(other, int) or isinstance(other, float):
            other = ReverseMode([other]*len(self.value))
        z = ReverseMode( -self.value + other.value)
        self.child.append((-np.ones(self.value.shape), z)) 
        other.child.append((np.ones(self.value.shape), z)) 
        return z

    def __mul__(self, other):
        """
        Overloads the multiplication operation
        Inputs: Scalar or ReverseMode Instance
        Returns: A new AutoDiff object which is the result of the multiplication operation
        performed between the AutoDiff object and the argument that was passed in

        Example:
        >>> x = ReverseMode([-1,-2])
        >>> y = ReverseMode([10,20])
        >>> z = x * y
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([10, 20])
        >>>
        >>> x = ReverseMode([-1,-2])
        >>> z = x * 3
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([3, 3])
        """
        if isinstance(other, int) or isinstance(other, float):
            other = ReverseMode([other]*len(self.value))
        z = ReverseMode(self.value * other.value)
        self.child.append((other.value, z)) # weight = dz/dself = other.value
        other.child.append((self.value, z)) # weight = dz/dother = self.value
        return z

    def __rmul__(self, other):
        """
        Inputs: Scalar or AutoDiff Instance
        Returns: A new AutoDiff object which is the result of the multiplication operation
        performed between the AutoDiff object and the argument that was passed in

        Example:
        >>> x = ReverseMode([-1,-2])
        >>> z = 2 * x
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-2, -2])
        """
        return self * other

    def __truediv__(self, other):
        """
        Overloads the division operation
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the ReverseMode
        object divided by the argument that was passed in

        Example:
        >>> x = ReverseMode([-1,-2])
        >>> y = ReverseMode([10,20])
        >>> z = x / y
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1. , 0.5])
        >>>
        >>> x = ReverseMode([-1,-2])
        >>> z = x / 3
        >>> z.gradient_coef = 2
        >>> x.get_derivative()
        array([0.66666667, 0.66666667])
        """
        return self * (other ** (-1))

    def __rtruediv__(self, other):
        """
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the argument that
        was passed in divided by the ReverseMode object

        Example:
        >>> x = ReverseMode([10,5])
        >>> z = 2 / x
        >>> z.gradient_coef = 2
        >>> x.get_derivative()
        array([-0.04, -0.16])
        """
        return other*(self**(-1))

    def __pow__(self, other):
        """
        Overloads the power operation
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the ReverseMode object being
        raised to the power of the argument that was passed in

        Example:
        >>> x = ReverseMode([-2,-3])
        >>> y = ReverseMode([1,2])
        >>> z = x ** y
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([ 1., -6.])
        >>>
        >>> x = ReverseMode([-2,-3])
        >>> z = x ** 3
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([ 12., 27.])
        """

        if isinstance(other, ReverseMode):
            if len(other.value) == 1:
                other_val = other.value*np.ones(self.value.shape)
            elif len(other.value) != len(self.value):
                raise ValueError("You must have two vectors of the same length to use power on both.")
            else:
                other_val = other.value[:]
            val = np.array([float(v) ** o for v, o in zip(self.value, other_val)])
            z = ReverseMode(val)
            temp_der = np.array([float(v) ** (o - 1) for v, o in zip(self.value, other_val)])
            self.child.append((other_val * temp_der, z)) # weight = dz/dself
            other.child.append((val*np.log(self.value), z))
            return z
        elif isinstance(other, (float, int)):
            val = np.array([float(v) ** other for v in self.value])
            z = ReverseMode(val)
            temp_der = np.array([float(v) ** (other - 1) for v in self.value])
            self.child.append((other * temp_der, z))
            return z
        else:
            raise TypeError("Please us power with another ReverseMode object, int or float.")

    def __rpow__(self, other):
        """
        Inputs: Scalar or ReverseMode Instance
        Returns: A new ReverseMode object which is the result of the argument that was
        passed in being raised to the power of the ReverseMode object

        Example:
        >>> x = ReverseMode([-2,-3])
        >>> z = -2 ** x
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-0.1732868, -0.0866434])
        """

        if isinstance(other, ReverseMode):
            if len(other.value) == 1:
                other_val = other.value*np.ones(self.value.shape)
            elif len(other.value) != len(self.value):
                raise ValueError("You must have two vectors of the same length to use power on both.")
            else:
                other_val = other.value[:]
            val = np.array([float(o) ** v for v, o in zip(self.value, other_val)])
            z = ReverseMode(val)
            temp_der = np.array([(v*(float(o) ** (v-1))) for v, o in zip(self.value, other_val)])
            other.child.append((temp_der, z))
            self.child.append((val*np.log(other.value), z))
            return z
        elif isinstance(other, (float, int)):
            val = np.array([float(other) ** v for v in self.value])
            z = ReverseMode(val)
            self.child.append((val*np.log(other), z))
            return z
        else:
            raise TypeError("Please us power with another ReverseMode object, int or float.")


    def __neg__(self):
        """
        Inputs: None
        Returns: A new ReverseMode object which has the signs of
        the value and derivative ReverseModed

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = -x
        >>> z.gradient_coef = -3
        >>> x.get_derivative()
        array([3, 3])
        """
        return self.__mul__(-1)

    def __pos__(self):
        """
        Inputs: None
        Returns: The ReverseMode instance itself

        Example:
        >>> x = ReverseMode([1,2])
        >>> z = +x
        >>> z.gradient_coef = -3
        >>> x.get_derivative()
        -3
        """
        return self

    def sin(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the sine computation done on the value and derivative

        Example:
        >>> x = ReverseMode([3,5])
        >>> z = ReverseMode.sin(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-0.9899925 ,  0.28366219])
        """
        z = ReverseMode(np.sin(self.value))
        self.child.append((np.cos(self.value), z)) # z = sin(x) => dz/dx = cos(x)
        return z

    def cos(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the cosine computation
        done on the value and derivative

        Example:
        >>> x = ReverseMode([3,5])
        >>> z = ReverseMode.cos(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-0.14112001,  0.95892427])
        """
        z = ReverseMode(np.cos(self.value))
        self.child.append((-np.sin(self.value), z))
        return z

    def tan(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the tangent computation
        done on the value and derivative

        Example:
        >>> x = ReverseMode([1,2])
        >>> z = ReverseMode.tan(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([3.42551882, 5.7743992 , 1.02031952])
        """
        z = ReverseMode(np.tan(self.value))
        self.child.append((1 / (np.cos(self.value) ** 2), z))
        return z

    def sinh(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the hyperbolic sine
        computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-1,-2])
        >>> z = ReverseMode.sinh(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1.54308063, 3.76219569])
        """
        z = ReverseMode(np.sinh(self.value))
        self.child.append((np.cosh(self.value), z))
        return z

    def cosh(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the hyperbolic cosine
        computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-1,2])
        >>> z = ReverseMode.cosh(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-1.17520119,  3.62686041])
        """
        z = ReverseMode(np.cosh(self.value))
        self.child.append((np.sinh(self.value), z))
        return z

    def tanh(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the hyperbolic
        tangent computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-1,2])
        >>> z = ReverseMode.tanh(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.41997434, 0.07065082])
        """
        z = ReverseMode(np.tanh(self.value))
        self.child.append((1 / (np.cosh(self.value) ** 2), z))
        return z

    def exp(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the natural exponential
        computation done on the value  and derivative

        Example:
        >>> x = ReverseMode([2,3])
        >>> z = ReverseMode.exp(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([ 7.3890561 , 20.08553692])

        """
        z = ReverseMode(np.exp(self.value))
        self.child.append((np.exp(self.value), z))
        return z

    def exp_base(self, base=np.exp(1)):
        """
        Inputs: scalar
        Returns: A new ReverseMode object with the exponential (using a specified base)
        computation done on the value and derivative

        Example:
        >>> x = ReverseMode([1,2])
        >>> z = ReverseMode.exp_base(x, 2)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1.38629436, 2.77258872])
        """
        if isinstance(base, (int, float)):
            return self.__rpow__(base)
        else:
            raise TypeError("Please enter an int or float for base.")


    def arcsin(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the inverse
        sine computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-0.1,-0.2])
        >>> z = ReverseMode.arcsin(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([1.00503782, 1.02062073])
        """
        z = ReverseMode(np.arcsin(self.value))
        self.child.append((1 / (1 - (self.value**2))**0.5, z))
        return z

    def arccos(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the inverse
        cosine computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-0.1,-0.2])
        >>> z = ReverseMode.arccos(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([-1.00503782, -1.02062073])
        """
        z = ReverseMode(np.arccos(self.value))
        self.child.append((-1 / (1 - (self.value**2))**0.5, z))
        return z

    def arctan(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the inverse
        cosine computation done on the value and derivative

        Example:
        >>> x = ReverseMode([-0.1,-0.2])
        >>> z = ReverseMode.arctan(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.99009901, 0.96153846])
        """
        z = ReverseMode(np.arctan(self.value))
        self.child.append((1 / (1 + (self.value**2)), z))
        return z

    def logistic(self):
        """
        Inputs: None
        Returns: A new ReverseMode object calculated with logistic function

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = ReverseMode.logistic(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([4.53958077e-05, 2.06115361e-09])
        """
        z = ReverseMode(1/(1+np.exp(-self.value)))
        self.child.append((np.exp(-self.value)/((1+np.exp(-self.value))**2), z)) 
        return z


    def sqrt(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the square root
        computation done on the value and derivative

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = ReverseMode.sqrt(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.15811388, 0.1118034 ])
        """
        z = ReverseMode(self.value ** (1 / 2))
        self.child.append(((1 / 2) * (self.value ** (- 1 / 2)), z))
        return z

    def ln(self):
        """
        Inputs: None
        Returns: A new ReverseMode object with the natural log
        computation done on the value and derivative

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = ReverseMode.ln(x)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.1 , 0.05])
        """
        if (np.all(self.value > 0)) :
            z = ReverseMode(np.log(self.value))
            self.child.append(( 1 / self.value , z))
            return z
        else:
            raise ValueError("ReverseMode.ln only takes in positive numbers as arguments.")


    def log(self, base=np.exp(1)):
        """
        Inputs: scalar
        Returns: A new ReverseMode object with the log (using a specified
        base) computation done on the value and derivative

        Example:
        >>> x = ReverseMode([10,20])
        >>> z = ReverseMode.log(x, 2)
        >>> z.gradient_coef = 1
        >>> x.get_derivative()
        array([0.1442695 , 0.07213475])
        """
        if (np.all(self.value > 0) and base > 0):
            z = ReverseMode(np.log(self.value) / np.log(base))
            self.child.append((1 / (self.value * np.log(base)), z))
            return z
        else:
            raise ValueError("ReverseMode.log only takes in positive numbers as arguments.")

    def __eq__(self, other):
        """
        Overloads the equal comparision operator (==)
        Inputs: ReverseMode Instance
        Returns: True if self and other ReverseMode instance have the
        same value; False if not

        Example:
        >>> v = ReverseMode([1,2])
        >>> w = ReverseMode([3,4])
        >>> x = ReverseMode([5,4])
        >>> y = ReverseMode([5,4s])
        >>> v == w
        array([False, False])
        >>> w == x
        array([False,  True])
        >>> x == y
        array([ True,  True])
        """
        if isinstance(other, (int, float)):
            return np.equal(self.value, other)
        elif isinstance(other, ReverseMode):
            return np.equal(self.value, other.value)
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")

    def __ne__(self, other):
        """
        Overloads the not equal comparision operator (!=)
        Inputs: ReverseMode Instance
        Returns: False if self and other ReverseMode instance have the
        same value; True if not

        Example:
        >>> v = ReverseMode([3,4])
        >>> w = ReverseMode([4,6])
        >>> x = ReverseMode([5,6])
        >>> y = ReverseMode([5,6])
        >>> v != w
        array([ True,  True])
        >>> w != x
        array([ True, False])
        >>> x != y
        array([False, False])
        """
        if isinstance(other, (int, float)):
            return not np.equal(self.value, other)
        elif isinstance(other, ReverseMode):
            return self.value != other.value
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")


    def __lt__(self, other):
        """
        Compares the value of the ReverseMode instance with the input
        Inputs: Scalar or ReverseMode Instance
        Returns: True if self.value is less than other.value or
        if self.value is less than other; False if not

        Example:
        >>> v = ReverseMode([3,4])
        >>> w = ReverseMode([4,6])
        >>> x = ReverseMode([5,6])
        >>> y = ReverseMode([5,6])
        >>> v < w
        array([ True,  True])
        >>> w < x
        array([ True, False])
        >>> x < y
        array([False, False])
        """
        if isinstance(other, (int, float)):
            return np.less(self.value, other)
        elif isinstance(other, ReverseMode):
            return np.less(self.value, other.value)
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")

    def __le__(self, other):
        """
        Compares the value of the ReverseMode instance with the input
        Inputs: Scalar or ReverseMode Instance
        Returns: True if self.value is less than or equal to other.value or
        if self.value is less than or equal to other; False if not

        Example:
        >>> v = ReverseMode([3,4])
        >>> w = ReverseMode([4,6])
        >>> x = ReverseMode([5,6])
        >>> y = ReverseMode([5,6])
        >>> w <= v
        array([False, False])
        >>> x <= w
        array([False,  True])
        >>> y <= x
        array([ True,  True])
        """
        if isinstance(other, (int, float)):
            return np.less_equal(self.value, other)
        elif isinstance(other, ReverseMode):
            return np.less_equal(self.value, other.value)
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")

    def __gt__(self, other):
        """
        Compares the value of the ReverseMode instance with the input
        Inputs: Scalar or ReverseMode Instance
        Returns: True if self.value is greater than other.value or
        if self.value is greater than other; False if not

        Example:
        >>> v = ReverseMode([3,4])
        >>> w = ReverseMode([4,6])
        >>> x = ReverseMode([5,6])
        >>> y = ReverseMode([5,6])
        >>> w > v
        array([ True,  True])
        >>> x > w
        array([ True, False])
        >>> y > x
        array([False, False])
        """
        if isinstance(other, (int, float)):
            return np.greater(self.value, other)
        elif isinstance(other, ReverseMode):
            return np.greater(self.value, other.value)
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")

    def __ge__(self, other):
        """
        Compares the value of the ReverseMode instance with the input
        Inputs: Scalar or ReverseMode Instance
        Returns: True if self.value is greater than or equal to other.value or
        if self.value is greater than or equal to other; False if not

        Example:
        >>> v = ReverseMode([3,4])
        >>> w = ReverseMode([4,6])
        >>> x = ReverseMode([5,6])
        >>> y = ReverseMode([5,6])
        >>> v >= w
        array([False, False])
        >>> w >= x
        array([False,  True])
        >>> x >= y
        array([ True,  True])
        """
        if isinstance(other, (int, float)):
            return np.greater_equal(self.value, other)
        elif isinstance(other, ReverseMode):
            return np.greater_equal(self.value, other.value)
        else:
            raise TypeError("Please only compare ReverseMode object with another ReverseMode object or int or float.")





class ReverseModeVector():
 

    # The ReverseModeVector class is the interface through which the user can set functions
    # and get the derivative of a function
    # Note: The output function can be a vector function, and each function can be defined by multiple variables
    # This means that any mapping from R^n to R^m is possible
    # To further simplify the usage of the package, each function can be simultaneously evaluated at numerous points
    # Therefore, the derivative of a multidimensional multivariate function evaluated at differnent points
    # will return a Tensor (!)

    def __init__(self, vector_function):
        """
        Initializes the ReverseModeVector object
        
        Inputs: A single function (ReverseMode object) or a list of functions (List of ReverseMode objects)
        Returns: None

        Example:
        >>> x = ReverseMode([1,3])
        >>> y = ReverseMode([2,4])
           
        >>> f1 = x**2
        >>> f2 = ReverseMode.exp(ReverseMode.sin(x*y))
        
        >>> rm = ReverseModeVector([f1,f2])
        """
        
        self.vector_function = []
        
        ## call the set_function if vector_function is not empty
        if vector_function:
            self.set_function(vector_function)
        
            
    def set_function(self,vector_function):
        """
        User can set multidimensional function which should be evaluated later
        
        Inputs: A single function (ReverseMode object) or a list of functions (List of ReverseMode objects)
        Returns: None

        Example:
        >>> x = ReverseMode([1,3])
        >>> y = ReverseMode([2,4])
           
        >>> f1 = x**2
        >>> f2 = ReverseMode.exp(ReverseMode.sin(x*y))
        
        >>> rm = ReverseModeVector()
        >>> rm.set_function([f1,f2])
        """
        
        if not vector_function:
            raise Exception("No function was specified")
        
        ## make it a list if it's only a single function and check if its the correct type
        if not isinstance(vector_function, list):
            if not isinstance(vector_function, ReverseMode):
                raise TypeError("A function has to be of type: ReverseMode")
            
            self.vector_function = [vector_function]
        
        ## if the vector_function is already a list, then check for correct type and set instance variable
        else:
            
            for f in vector_function:
                if not isinstance(f, ReverseMode):
                    raise TypeError("A function has to be of type: ReverseMode")
            
            
            self.vector_function = vector_function


    def get_derivative(self, input_values):
        """
        Evaluates the derivative of the specified function at numerous points simultaneously
        
        Inputs: A single variable (ReverseMode object) or a list of variables (List of ReverseMode objects)
        Returns: Tensor of derivative evaluated at all specified points

        Example:
        >>> x = ReverseMode([1,3])
        >>> y = ReverseMode([2,4])
           
        >>> f1 = x**2
        >>> f2 = ReverseMode.exp(ReverseMode.sin(x*y))
        
        >>> rm = ReverseModeVector([f1,f2])
        >>> rm.get_derivative([x,y])
        [[[ 2.         -2.06623374]
          [ 0.         -1.03311687]]

         [[ 6.          1.97377037]
          [ 0.          1.48032778]]]
        
        ### Explanation ###
        The first matrix is the Jacobian of the multidimenional function evaluated at (x,y) = (1,2)
        The second matrix is the Jacobian of the function evaluated at the second point (x,y) = (3,4)
        
        """
        
        #if no function, throw error
        if not self.vector_function:
            raise Exception("No function was specified")
         
        #if no variables, throw error
        if not input_values:
            raise Exception("No variables were specified")
        
        output_tensor = []
        
        if not isinstance(input_values, list):
            input_values_temp = [[input_values]]
        
        else:
            input_values_temp = [input_values]
        
        input_values = input_values_temp

        
        for i, inp_val in enumerate(input_values):
        
            # output_tensor.append([])
            # for val in input_values[i]:
            for val in inp_val:
                temp_derivative = []
                for j in range(len(self.vector_function)):
                    # temp_derivative = []
                    val.reset_gradient()
                    self.vector_function[j].gradient_coef = 1
                    derivative_value = val.get_derivative()
                    input_values = input_values_temp
                    temp_derivative.append(derivative_value)
                output_tensor.append(temp_derivative)
                    # input_values = input_values_temp
                    
        #this will become the final output array           
        outin = np.array(output_tensor)
        outin = outin.T
        
        #print(outin.shape)
        
        if len(outin.shape) == 3:
            outv = np.empty((outin.shape[0], outin.shape[2], outin.shape[1]))
            
        if len(outin.shape) == 2:
            outv = np.empty((outin.shape[0], outin.shape[1]))
        
        #print(outv.shape)
        for i in range(outv.shape[0]):
            outv[i] = outin[i].T
        
        
        return np.array(outv)
    
    
    def get_function_value(self):
        """
        User can evaluate the value of the function at numerous points simultaneously
        
        Inputs: None (ReverseModeVector has access through the functions through the self.vector_function variable)
        Returns: Matrix of function values (first row is function evaluated at first point, second row function evaluated at second point etc)

        Example:
        >>> x = ReverseMode([1,3,12])
        >>> y = ReverseMode([2,4,0])
           
        >>> f1 = x**2
        >>> f2 = 2*x+y**2
    
        >>> rm = ReverseModeVector([f1,f2])
    
        >>> rm.get_function_value()
        [[  1.   6.]  # function evaluated at (1,2)
         [  9.  22.]  # function evaluated at (3,4)
         [144.  24.]]  # function evaluated at (12,0)
        """
        
        if not self.vector_function:
            raise Exception("No function was specified")
        
        vals = np.array([f.value for f in self.vector_function])
        
        return vals.T








# -*- coding: utf-8 -*-

import numpy as np

class DualNumber:
    
    def __init__(self, real = 1.0, dual = np.array([1.0])):
        """
        __init__(self, real = 1.0, dual = np.array([1.0]))

        Initiate a DualNumber object.
         
        Parameters
        ----------
        self : DualNumber object
        real : integer/float, optional
            The value of the variable.
            Defalut value is 1.0.
        dual : array, optional
            The initial derivative of the variable.
            Default value is np.array([1.0]).
         
        Examples
        --------
        >>> x = DualNumber()

        Neither input real and dual value is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        """
        self.real = float(real)
        self.dual = dual       
  

    def __add__(self, other):
        """
        __add__(self, other)

        Dunder method for adding anthor DualNumber or number.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `other` added.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = x + 1.2
        >>> x.real
        1.2
        >>> x.dual
        array([1.])

        Neither adding a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = x + y
        >>> z.real
        5.0
        >>> z.dual
        array([1, 1])
        """
        # if other is also a dual number
        if(isinstance(other, DualNumber)):
                       
            return DualNumber(self.real+other.real, self.dual+other.dual)
            
        # if other is an int or a float
        elif(isinstance(other, (int, float))):
                   
            return DualNumber(self.real+other, self.dual)
        
        # if trying to add an unsupported thing to a dual number
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float!')
    

    def __sub__(self, other):
        """
        __sub__(self, other)

        Dunder method for substracting anthor DualNumber or number.
    
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `other` subtracted.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = x - 1.2
        >>> x.real
        -1.2
        >>> x.dual
        array([1.])

        Neither substracting a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = x - y
        >>> z.real
        -1.0
        >>> z.dual
        array([ 1, -1])
        """
        # reuse the add method
        # Note: if other is a dual number, this will automatically call the __mul__ method of that instance
        return self.__add__(-1*other) 


    def __rsub__(self, other):
        """
        __rsub__(self, other)

        Dunder method for substacted by anthor DualNumber or number.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with substracted by `other`.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = 1 - x
        >>> x.real
        1.0
        >>> x.dual
        array([-1.])

        Neither subtracted by a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = y - x
        >>> z.real
        1.0
        >>> z.dual
        array([-1,  1])
        """
        ##reuse the add method; Note: if other is a dual number, this will automatically call the __mul__ method of that instance
        t = self.__mul__(-1)
        return t.__add__(other)


    def __neg__(self):
        """
        __neg__(self)

        Dunder method for taking the negative.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object negation.
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = - x
        >>> x.real
        -0.0
        >>> x.dual
        array([-1.])
        """
        return DualNumber(-self.real, -self.dual)

    __isub__ = __sub__   
        
   
    def __mul__(self, other):
        """
        __mul__(self, other)

        Dunder method for multiplying anthor DualNumber or number.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `other` multiplied.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = x * 4
        >>> x.real
        0.0
        >>> x.dual
        array([4.])

        Neither multiplying a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = x + y
        >>> z.real
        6.0
        >>> z.dual
        array([3., 2.])
        """
        #if other is also a dual number
        if(isinstance(other, DualNumber)):
             return DualNumber(self.real*other.real, self.real*other.dual+self.dual*other.real)
        
        #if other is an int or a float
        elif(isinstance(other, (int, float))):   
            return DualNumber(self.real*other, self.dual*other)
        
        #if trying to add an unsupported thing to a dual number
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float!')
    
    
    # when saying like dual number += sth, then it is the same as dual number + sth (but python calls a different dunder method for the += case)
    __iadd__ = __add__
    __imul__ = __mul__
     
    #saying that reverse multiplication is the same as multiplication   
    __rmul__ = __mul__  
    #saying that reverse addition is the same as addition   
    __radd__ = __add__  


    def __floordiv__(self, *args):
        """
        Raises
        ------
        NotImplementedError
          Floor division doesn't really make sense in our use case.
          Therefore raise an exception.
        """
        raise NotImplementedError('Floor division is not supported!')


    __rfloordiv__ = __floordiv__


    def __ifloordiv__(self, *args):
        """
        Raises
        ------
        NotImplementedError
          Ifloor division doesn't really make sense in our use case.
          Therefore raise an exception.
        """
        raise NotImplementedError('Floor division is not supported!')

       
    def __truediv__(self, other):
        """
        __truediv__(self, other)

        Dunder method for dividing anthor DualNumber or number.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `other` divided.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = x / 2
        >>> x.real
        0.0
        >>> x.dual
        array([0.25])

        Neither dividing a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = x / y
        >>> z.real
        0.6666666666666666
        >>> z.dual
        array([ 0.33333333, -0.22222222])
        """
        #calculate the division complement of the other object, then reuse multiplication
        #e.g. if other = z1 (dual number), then calculate '1/z1' and reuse the mul-dunder-method
        
        #if other is also a dual number
        if(isinstance(other, DualNumber)):
            
            t = DualNumber(1/other.real, -1*(other.dual/(other.real*other.real)))
            
            return self.__mul__(t)
        
        #if other is an int or a float
        elif(isinstance(other, (int, float))):
            ##reusing the method from above
            
            t = DualNumber(1/other, 0)
            
            return self.__mul__(t)
        
        #if trying to add an unsupported thing to a dual number
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float!')


    #re-implementing the reverse true division of dual numbers, because division is NOT commutative
    def __rtruediv__(self, other):
        """
        __add__(self, other)

        Dunder method for being divided by anthor DualNumber or number.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object divided by `other`.

        Raises
        ------
        AttributeError
          If `other` is not integer, float or DualNumber. 
  
        Examples
        --------
        >>> x = DualNumber(real = 1)
        >>> x = 1 / x
        >>> x.real
        1.0
        >>> x.dual
        array([-1.])

        Neither divided by a DualNumber is acceptable.

        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> z = y / x
        >>> z.real
        1.5
        >>> z.dual
        array([-0.75,  0.5 ])
        """
        #we are creating the divisional complement of the instance of self, then we multiply it with other      
        t = DualNumber(1/self.real, -1*(self.dual/(self.real*self.real)))
            
        return t.__mul__(other)
        
    
    #to support both the __truediv__ as well as __div__ methods, which are the same in our context
    __div__ = __truediv__
    __itruediv__ = __truediv__
    

    def __eq__(self, other):
        """
        __eq__(self, other)

        Dunder method for comparing self with other to check if they are equal.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber

        Returns
        -------
        output : bool, True/False
          If self.real=other.real and self.dual=other.dual, retrurn True else return False.

        Examples
        --------
        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> x == y
        False
        """
        return self.real==other.real and (self.dual==other.dual).all()

    def __ne__(self, other):
        """
        __add__(self, other)

        Dunder method for comparing self ith other to check if they are not equal.
         
        Parameters
        ----------
        self : DualNumber object
        other : DualNumber/integer/float

        Returns
        -------
        output : bool, True/False
          If self.real!=other.real or self.dual=other.dual, retrurn True else return False.
  
        Examples
        --------
        >>> x = DualNumber(real = 2, dual = np.array([1,0]))
        >>> y = DualNumber(real = 3, dual = np.array([0,1]))
        >>> x != y
        True
        """
        return self.real!=other.real or (self.dual!=other.dual).any()


    def __lt__ (self, other):
        """
        __lt__(self, other)
        
        Comapring self with other to check if self is less than other.

        Raises
        ------
        NotImplementedError
          Less than doesn't really make sense in our use case.
          Therefore raise an exception.
        """
        raise NotImplementedError('Less than comparison is not appliable to Dual Numbers!')


    def __gt__ (self, other):
        """
        __gt__(self, other)
        
        Comapring self with other to check if self is greater than other.

        Raises
        ------
        NotImplementedError
          Greater than doesn't really make sense in our use case.
          Therefore raise an exception.
        """
        raise NotImplementedError('Greater than comparison is not appliable to Dual Numbers!')


    def __le__ (self, other):
        """
        __le__(self, other)
        
        Comapring two dual numbers if self is less than or euqal to other.

        Raises
        ------
        NotImplementedError
          Less than doesn't really make sense in our use case.
          Therefore raise an exception.
        """
        raise NotImplementedError('Less than or equal to comparison is not appliable to Dual Numbers!')

    def __ge__ (self, other):
        """
        __ge__(self, other)
        
        Comapring self with other to check if self is greater than other.

        Raises
        ------
        NotImplementedError
          Greater than doesn't really make sense in our use case
          Therefore raise an exception
        """
        raise NotImplementedError('Greater than or equal comparison is not appliable to Dual Numbers!')
    

    def __pow__(self, p):
        """
        __pow__(self, p)

        Dunder method for powering self to p.
         
        Parameters
        ----------
        self : DualNumber object
        p : DualNumber or integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `p` powered.

        Raises
        ------
        AttributeError
          If p is not a DualNumber, integer or float.
  
        Examples
        --------
        >>> x = DualNumber(real = 2)
        >>> x = x**3.4
        >>> x.real
        10.556063286183154
        >>> x.dual
        array([17.94530759])

        Neither DualNumber for p is acceptable.

        >>> x = DualNumber(real = 1.4, dual = np.array([1,0]))
        >>> y = DualNumber(real = 2.5, dual = np.array([0,1]))
        >>> z = x ** y
        >>> z.real
        2.319103274975049
        >>> z.dual
        array([4.14125585, 0.78031387])
        """
        if(isinstance(p, DualNumber)):
            return DualNumber(np.power(self.real,p.real), np.power(self.real,p.real)*np.log(self.real)*p.dual+p.real*np.power(self.real,p.real-1)*self.dual)
        elif(isinstance(p, (int, float))):
            return DualNumber(np.power(self.real,p), p*np.power(self.real,p-1)*self.dual)
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float')


    def __rpow__(self, p):
        """
        __rpow__(self, p)

        Dunder method for powering p to self.
         
        Parameters
        ----------
        self : DualNumber object
        p : DualNumber or integer/float

        Returns
        -------
        output : DualNumber
             A DualNumber object with `self` powered.

        Raises
        ------
        AttributeError
          If p is not a DualNumber, integer or float.
  
        Examples
        --------
        >>> x = DualNumber(real = 1.4)
        >>> x = 2.3 ** x
        >>> x.real
        3.2093639532679714
        >>> x.dual
        array([2.67310852])

        Neither DualNumber for p is acceptable.

        >>> x = DualNumber(real = 1.4, dual = np.array([1,0]))
        >>> y = DualNumber(real = 2.5, dual = np.array([0,1]))
        >>> z = x ** y
        >>> z.real
        2.319103274975049
        >>> z.dual
        array([0.78031387, 4.14125585])
        """
        return self.exp(base = p)



    def exp(self, base=np.exp(1)):
        """
        exp(self, base=np.exp(1))

        Taing exponential for any base.
  
        Parameters
        ----------
        self : DualNumber object
        base : DualNumber or integer/float, optional 
            The base of the exponential.
            Default value of the base is e.

        Returns
        -------
        output : DualNumber
             A DualNumber object with exponential for the base. 

        Raises
        ------
        AttributeError
          If p is not a DualNumber, integer or float.
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = DualNumber.exp(x)
        >>> x.real
        1.0
        >>> x.dual
        array([1.])

        Neither DualNumber for base is acceptable.

        >>> x = DualNumber(real = 1.4, dual = np.array([1,0]))
        >>> y = DualNumber(real = 2.5, dual = np.array([0,1]))
        >>> z = DualNumber.exp(x, base = y)
        >>> z.real
        3.6067497647680336
        >>> z.dual
        array([3.30483138, 2.01977987])
        """
        if(isinstance(base, DualNumber)):
            return DualNumber(np.power(base.real,self.real), np.power(base.real,self.real)*np.log(base.real)*self.dual+self.real*np.power(base.real,self.real-1)*base.dual)
        elif(isinstance(base, (int, float))):
            return DualNumber(np.power(base,self.real), np.power(base,self.real)*np.log(base)*self.dual)
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float')


    def sin(self):
        """
        sin(self)

        Taking sin.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with sin.
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = np.sin(x)
        >>> x.real
        0.0
        >>> x.dual
        array([1.])
        """
        return DualNumber(np.sin(self.real),self.dual*np.cos(self.real))
    

    def cos(self):
        """
        sin(self)

        Taking cos.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with cos.
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = np.cos(x)
        >>> x.real
        1.0
        >>> x.dual
        array([-0.])
        """
        return DualNumber(np.cos(self.real),self.dual*np.sin(self.real)*(-1))
    

    def tan(self):
        """
        sin(self)

        Taking tan.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with tan.
             
        Raises
        ------
        ValueError
          If self.real == pi/2 + (pi * n), where n is an integer.  
  
        Examples
        --------
        >>> x = DualNumber(real = 0)
        >>> x = np.tan(x)
        >>> x.real
        0.0
        >>> x.dual
        array([1.])
        """
        if self.real % np.pi == (np.pi/2):
            raise ValueError('Cannot take the tangent of pi/2 + (pi * n) when n is an integer!')
        return DualNumber(np.tan(self.real), self.dual*1/(np.cos(self.real)**2))
    

    def sqrt(self):
        """
        sqrt(self)

        Taking sqrt.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with sqrt.

        Raises
        ------
        ValueError
          If self.real < 0.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = np.sqrt(x)
        >>> x.real
        1.0
        >>> x.dual
        array([0.5])
        """
        if self.real < 0:
            raise ValueError('Cannot take sqrt of a negative value!')
        return DualNumber(np.sqrt(self.real), self.dual*1/(2*np.sqrt(self.real)))
    

    def log(self, base=np.exp(1)):
        """
        log(self, base=np.exp(1))

        Taking log for any base.
         
        Parameters
        ----------
        self : DualNumber object
        base : DualNumber, integer/float, optional 
            The base of the log function.
            Default value for the base is e. 

        Returns
        -------
        output : DualNumber
             A DualNumber object with taking log.

        Raises
        ------
        ValueError
          If self.real <= 0 or base/base.real <= 0.
        AttributeError
          If base is not a DualNumber, integer or float.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = DualNumber.log(x)
        >>> x.real
        0.0
        >>> x.dual
        array([1.])

        Neither DualNumber for base is acceptable.

        >>> x = DualNumber(real = 1.4, dual = np.array([1,0]))
        >>> y = DualNumber(real = 2.5, dual = np.array([0,1]))
        >>> z = DualNumber.log(x, base = y)
        >>> z.real
        0.3672112190123348
        >>> z.dual
        array([ 0.77954048, -0.16030336])
        """
        if self.real <= 0:
            raise ValueError('Cannot take the log of a nonpositive number!')
        if(isinstance(base, DualNumber)):
            if base.real <= 0:
                raise ValueError('Cannot take the nonpositive base!')
            return DualNumber(np.log(self.real)/np.log(base.real), self.dual/(self.real*np.log(base.real))-base.dual*np.log(self.real)/(base.real*(np.log(base.real)**2)))
        elif(isinstance(base, (int, float))):
            if base <= 0:
                raise ValueError('Cannot take the nonpositive base!')
            return DualNumber(np.log(self.real)/np.log(base), self.dual/(self.real*np.log(base)))
        else:
            raise AttributeError('Attribute not supported, e.g. not a dual number or int / float')


    def arcsin(self):
        """
        arcsin(self)

        Taking arcsin.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with arcsin.

        Raises
        ------
        ValueError
          If self.real <= -1 or self.real >= 1.
  
        Examples
        --------
        >>> x = DualNumber(real = 0.5)
        >>> x = np.arcsin(x)
        >>> x.real
        0.5235987755982989
        >>> x.dual
        array([1.15470054])
        """
        if self.real >= 1 or self.real <= -1:
            raise ValueError('Please input value -1<x<1 for arcsin!')
        return DualNumber(np.arcsin(self.real), self.dual*1/(1-self.real**2)**(1/2))
    

    def arccos(self):
        """
        arccos(self)

        Taking log10.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with arccos.

        Raises
        ------
        ValueError
          If self.real <= -1 or self.real >= 1.
  
        Examples
        --------
        >>> x = DualNumber(real = 0.5)
        >>> x = np.arccos(x)
        >>> x.real
        1.0471975511965979
        >>> x.dual
        array([-1.15470054])
        """
        if self.real >= 1 or self.real <= -1:
            raise ValueError('Please input value -1<x<1 for arccos!')
        return DualNumber(np.arccos(self.real), self.dual*-1/(1-self.real**2)**(1/2))
    

    def arctan(self):
        """
        arctan(self)

        Taking arctan.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with arctan.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = np.arctan(x)
        >>> x.real
        0.7853981633974483
        >>> x.dual
        array([0.5])
        """
        return DualNumber(np.arctan(self.real), self.dual*1/(1+self.real**2))
    

    def sinh(self):
        """
        sinh(self)

        Taking sinh.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with sinh.

        Raises
        ------
        ValueError
          If self.real <= 0.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = np.sinh(x)
        >>> x.real
        1.1752011936438014
        >>> x.dual
        array([1.54308063])
        """
        return DualNumber(np.sinh(self.real), self.dual*np.cosh(self.real))


    def cosh(self):
        """
        cosh(self)

        Taking cosh.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with cosh.

        Raises
        ------
        ValueError
          If self.real <= 0.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = np.cosh(x)
        >>> x.real
        1.5430806348152437
        >>> x.dual
        array([1.17520119])
        """
        return DualNumber(np.cosh(self.real), self.dual*np.sinh(self.real))
    

    def tanh(self):
        """
        tanh(self)

        Taking tanh.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with tanh.

        Raises
        ------
        ValueError
          If self.real <= 0.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = np.tanh(x)
        >>> x.real
        0.7615941559557649
        >>> x.dual
        array([0.41997434])
        """
        return DualNumber(np.tanh(self.real), self.dual*1/(np.cosh(self.real)**2))


    def logistic(self):
        """
        logistic(self)

        Taking logistic.
         
        Parameters
        ----------
        self : DualNumber object

        Returns
        -------
        output : DualNumber
             A DualNumber object with logistic.
  
        Examples
        --------
        >>> x = DualNumber()
        >>> x = DualNumber.logistic(x)
        >>> x.real
        0.7310585786300049
        >>> x.dual
        array([0.19661193])
        """
        return DualNumber(np.exp(self.real)/(1 + np.exp(self.real)), self.dual*np.exp(-self.real)/(1+np.exp(-self.real))**2)




class ForwardMode():

    def __init__(self, *args):
        """
        __init__(self, *args)

        Initiate a ForwardMode object.
         
        Parameters
        ----------
        self : ForwardMode object
        *args : list, optional
             A list of functions that users want to get derivative and value.
         
        Examples
        --------
        >>> f = ForwardMode()

        Neither a list of functions is acceptable:

        >>> x = DualNumber(real = 1)
        >>> f = ForwardMode([x**2]) 
        """

        if len(args) > 0:
            self.f = args[0]
            self.number_of_funtions = len(self.f)
    
    def set_function(self, f):
        """
        set_function(self, f):

        Set the function that users want to get derivative and value.

        Parameters
        ----------
        self : ForwardMode object
        f : list
           A list of functions that users want to get derivative and value.
         
        Examples
        --------
        >>> x = DualNumber(real = 1)
        >>> f = ForwardMode()
        >>> f.set_function([x**2, np.exp(x)])
        """
        self.f = f
        self.number_of_funtions = len(self.f)
        
    def get_derivative(self):
        """
        get_derivative(self, f):

        Get the derivative/Jacobian matrix of the function.

        Parameters
        ----------
        self : ForwardMode object

        Returns
        -------
        output : array
            Returns the derivative/Jacobian matrix of the function.
         
        Examples
        --------
        >>> x = DualNumber(real = 1)
        >>> f = ForwardMode()
        >>> f.set_function([((x**2)+3.5*x-20)/x, x])
        >>> f.get_derivative()
        array([[21.],
            [1.]])
        """
        self.derivative = np.array([self.f[0].dual])
        for i in range(1, self.number_of_funtions):
            self.derivative = np.r_[self.derivative, np.array([self.f[i].dual])]
        return self.derivative
    
    def get_function_value(self):
        """
        get_derivative(self, f):

        Get the value of the function.

        Parameters
        ----------
        self : ForwardMode object

        Returns
        -------
        output : array
            Returns the value of the function.
         
        Examples
        --------
        >>> x = DualNumber(real = 1)
        >>> f = ForwardMode()
        >>> f.set_function([((x**2)+3.5*x-20)/x, x])
        >>> f.get_function_value()
        array([-15.5,   1. ])
        """
        return np.array([f.real for f in self.f])
        